from typing import Optional
from .shingleset import shingleset_k
from .wildcard import wildcard_shinglesets


def jaccard(A: set, B: set) -> float:
    """Jaccard Similarity Coefficient or Intersection over Union

    Parameters:
    -----------
    A, B : set
        Sets of unique shingles

    Return:
    -------
    metric : float
        The Jaccard Similarity Coefficient
    """
    u = float(len(A.intersection(B)))
    return u / (len(A) + len(B) - u)


def jaccard_strings(s1: str, s2: str,
                    k: Optional[int] = 1,
                    n_max_wildcards: Optional[int] = None
                    ) -> float:
    """Jaccard Similarity Coefficient between two shingled strings

    Parameters:
    -----------
    s1, s2 : str
        Strings to compare

    k : int (Default 1)
        k parameter for k-shingling

    n_max_wildcards : int
        Maximum number of wildcard characters per word.

    Return:
    -------
    metric : float
        The Jaccard Similarity Coefficient between shingled strings
    """
    # limit k to the shortest str len
    k_max = min(k, len(s1), len(s2))
    # Shingling
    A = shingleset_k(s1, k_max)
    B = shingleset_k(s2, k_max)
    # Wildcard characters
    if n_max_wildcards:
        A = A.union(wildcard_shinglesets(A, n_max_wildcards))
        B = B.union(wildcard_shinglesets(B, n_max_wildcards))
    # compute similarity
    return jaccard(set(A), set(B))



import kshingle as ks


def test1():
    # check length
    data = ['abc d abc de abc def', 'abc defg abc def gh abc def ghi']
    shingled = [ks.shingleseqs_k(s, k=9) for s in data]
    for n in range(1, 15):
        VOCAB = ks.identify_vocab(shingled, n_max_vocab=n)
        assert len(VOCAB) == n


def test2():
    data = ['a', 'ab']
    shingled = [ks.shingleseqs_k(s, k=2) for s in data]
    VOCAB = ks.identify_vocab(
        shingled, sortmode='prefer-shorter', n_min_count=1, n_max_vocab=None)
    assert VOCAB == ['a', 'b', 'ab']


def test3():
    data = ['a', 'ab']
    shingled = [ks.shingleseqs_k(s, k=2) for s in data]
    VOCAB = ks.identify_vocab(
        shingled, sortmode='times-length', n_min_count=1, n_max_vocab=None)
    assert VOCAB == ['a', 'ab', 'b']


def test4():
    data = ['a', 'ab']
    shingled = [ks.shingleseqs_k(s, k=2) for s in data]
    VOCAB = ks.identify_vocab(
        shingled, sortmode='sqrt-x-length', n_min_count=1, n_max_vocab=None)
    assert VOCAB == ['ab', 'a', 'b']


def test5():
    data = ['a', 'ab']
    shingled = [ks.shingleseqs_k(s, k=2) for s in data]
    VOCAB = ks.identify_vocab(
        shingled, sortmode='log-x-length', n_min_count=1, n_max_vocab=None)
    assert VOCAB == ['ab', 'a', 'b']



import kshingle as ks


def test1():
    A = {'a', 'b', 'ab'}
    B = {'a'}
    score = ks.jaccard(A, B)
    assert score == 1 / 3


def test2():
    k = 1
    A = ks.shingleset_k("hamsterkäufe", k)
    B = ks.shingleset_k("hamsterkauf", k)
    score = ks.jaccard(A, B)
    assert 0.909 < score < 0.910


def test3():
    k = 8
    A = ks.shingleset_k("hamsterkäufe", k)
    B = ks.shingleset_k("hamsterkauf", k)
    score = ks.jaccard(A, B)
    assert 0.448 < score < 0.449


def test4():
    k = 8
    A = ks.shingleset_k("hamsterkäufe", k)
    A = A.union(ks.wildcard_shinglesets(A, n_max_wildcards=1))
    B = ks.shingleset_k("hamsterkauf", k)
    B = B.union(ks.wildcard_shinglesets(B, n_max_wildcards=1))
    score = ks.jaccard(A, B)
    assert 0.412 < score < 0.413


def test5():
    s1 = "hamsterkäufe"
    s2 = "hamsterkauf"
    k = 8
    score = ks.jaccard_strings(s1, s2, k)
    assert 0.448 < score < 0.449


def test6():
    s1 = "hamsterkäufe"
    s2 = "hamsterkauf"
    k = 8
    n_max_wildcards = 1
    score = ks.jaccard_strings(s1, s2, k, n_max_wildcards)
    assert 0.412 < score < 0.413



import ray
import psutil
import random
from .cews import expandshingle
from typing import Optional, Dict
import numpy as np


@ray.remote
def ray_expandshingle(s: str,
                      db: Dict[str, int],
                      memo: Optional[dict],
                      wildcard: Optional[str] = '\uFFFF',
                      max_wildcards: Optional[int] = 1,
                      min_samples_split: Optional[int] = 2,
                      min_samples_leaf: Optional[int] = 1,
                      threshold: Optional[float] = 0.8):
    """ Ray.io wrapper for ks.expandshingle """
    return expandshingle(
        s, db=db, memo=memo,
        wildcard=wildcard,
        max_wildcards=max_wildcards,
        min_samples_split=min_samples_split,
        min_samples_leaf=min_samples_leaf,
        threshold=threshold)


def cews_cpu(db: Dict[str, int],
             memo: Optional[dict] = {},
             wildcard: Optional[str] = '\uFFFF',
             max_wildcards: Optional[int] = 1,
             min_samples_split: Optional[int] = None,
             min_samples_leaf: Optional[int] = 1,
             threshold: Optional[float] = 0.8,
             priority: Optional[str] = 'common',
             vocab_size: Optional[int] = None):
    # Start ray.io
    NUM_CPUS = max(1, int(psutil.cpu_count(logical=True) * 0.9))
    ray.init(num_cpus=NUM_CPUS)
    print(f"Num CPUs: {NUM_CPUS}")

    # add single-chars to memo automatically
    if len(memo) == 0:
        memo = {k: v for k, v in db.items() if len(k) == 1}

    # sort shingles by frequency
    db_list = list(db.items())
    db_list = sorted(db_list, key=lambda item: -item[1])

    # approximate `min_samples_leaf` with `vocab_size`
    if (min_samples_leaf == 'auto') and (vocab_size is not None):
        idx = min(len(db_list) - 1, vocab_size)
        min_samples_leaf = int(max(1, db_list[idx][1]))

    # if `min_samples_leaf` is float
    if isinstance(min_samples_leaf, float) and (0.0 < min_samples_leaf <= 1.0):
        cnts = np.array([c for _, c in db_list])
        mask = (cnts / np.sum(cnts)) > min_samples_leaf
        if mask.any():
            min_samples_leaf = int(cnts[mask][-1])
        else:
            min_samples_leaf = 1

    # check threshold
    assert threshold is not None
    assert (0.0 < threshold <= 1.0)

    # set `min_samples_split`
    if min_samples_split is None:
        min_samples_split = max(2, int(min_samples_leaf / threshold))

    # only shingles with at least `min_samples_leaf` counts
    if priority == 'rare':
        shingles = [s for s, _ in db_list]
        shingles.reverse()
    elif (priority == 'common') and (vocab_size is not None):
        shingles = [s for s, c in db_list if c >= min_samples_leaf]
    else:  # priority == 'common'
        shingles = [s for s, _ in db_list]

    # start parallel searches
    for start in range(0, len(shingles), NUM_CPUS):
        # run trees in parallel
        tmp_memos = []
        for s in shingles[start:(start + NUM_CPUS)]:
            tmp_memos.append(ray_expandshingle.remote(
                s, db=db, memo=memo,
                wildcard=wildcard,
                max_wildcards=max_wildcards,
                min_samples_split=min_samples_split,
                min_samples_leaf=min_samples_leaf,
                threshold=threshold))
        # collect results
        tmp_memos = ray.get(tmp_memos)
        for m in tmp_memos:
            memo.update(m)
        # early stopping
        if vocab_size is not None:
            if len(memo) >= vocab_size:
                break

    # stop ray
    ray.shutdown()
    # done
    return memo



import kshingle as ks
import pytest
import functools
import itertools
from collections import Counter


def test1():
    db = {"ab": 3, "ac": 4, "ad": 5}
    shingle = "a"
    memo = ks.expandshingle(
        shingle, db=db, memo={}, wildcard="?", threshold=1.0,
        min_samples_split=1, max_wildcards=0)
    assert memo == {}


def test2():
    db = {"ab": 3, "ac": 4, "ad": 5}
    shingle = "a"
    memo = ks.expandshingle(
        shingle, db=db, memo={}, wildcard="?", threshold=1.0,
        min_samples_split=1, max_wildcards=1)
    assert memo == {'a?': 3, 'ac': 4, 'ad': 5}


def test3():
    with pytest.raises(Exception):
        ks.expandshingle("a", db={}, memo={}, min_samples_split=0)


def test4():
    db = {"qar": 2, "qbr": 3, "qcr": 4, "qdr": 5, "qer": 6, "qfr": 7}
    shingle = "qar"
    memo = ks.expandshingle(
        shingle, db=db, memo={}, wildcard="?", threshold=.8,
        min_samples_split=1, max_wildcards=1)
    assert memo == {'q?r': 9, 'qdr': 5, 'qer': 6, 'qfr': 7}


def test4b():
    db = {"qar": 2, "qbr": 3, "qcr": 4, "qdr": 5, "qer": 6, "qfr": 7}
    memo = ks.cews(
        db, wildcard="?", threshold=.8, min_samples_split=1, max_wildcards=1)
    assert memo == {'q?r': 9, 'qdr': 5, 'qer': 6, 'qfr': 7}


def test5():
    db = {"ab": 2, "ac": 3, "abq": 4, "abr": 5}
    shingle = "ab"
    memo = ks.expandshingle(
        shingle, db=db, memo={}, wildcard="?", threshold=1,
        min_samples_split=1, max_wildcards=2)
    assert memo == {'a?r': 5, 'ab?': 4, 'abr': 5}


def test6():
    db = {"ab": 2, "abq": 4, "xabq": 2, "yabq": 3}
    shingle = "ab"
    memo = ks.expandshingle(
        shingle, db=db, memo={}, wildcard="?", threshold=1,
        min_samples_split=1, max_wildcards=2)
    assert memo == {'ab?': 4, '?ab?': 2, 'y??q': 3,
                    'y?bq': 3, 'ya?q': 3, 'yabq': 3}


# def test6b():
#     db = {"ab": 2, "abq": 4, "xabq": 2, "yabq": 3}
#     memo = ks.cews(
#         db, wildcard="?", threshold=.8, min_samples_split=3, max_wildcards=3)
#     assert memo == {'ab?': 4, 'a?q': 4, 'y?bq': 3, 'y??q': 3, 'ya?q': 3}


def test7():
    k = 5
    docs = [
        "Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam ",
        "nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam ",
        "erat, sed diam voluptua. At vero eos et accusam et justo duo ",
        "dolores et ea rebum. Stet clita kasd gubergren, no sea takimata "]
    # generate all shingles
    shingled = [ks.shingleseqs_k(doc, k=k) for doc in docs]
    assert len(shingled) == len(docs)
    assert len(shingled[0]) == k
    # run CEWS algorithm
    db = functools.reduce(lambda x, y: x + Counter(itertools.chain(*y)),
                          shingled, Counter([]))
    memo = ks.cews(db, threshold=0.8, min_samples_split=10, max_wildcards=2)
    # encode shingles with patterns
    PATTERNS = ks.shingles_to_patterns(memo)
    encoded = ks.encode_with_patterns(shingled, PATTERNS, len(PATTERNS))
    assert len(PATTERNS) == len(memo)
    assert len(encoded) == len(shingled)
    for i in range(len(encoded)):
        assert len(encoded[i]) == len(shingled[i])
        for j in range(len(encoded[i])):
            assert len(encoded[i][j]) == len(shingled[i][j])


def test8():
    memo = {"*a": 1, "a*b": 1, "a*": 1, "a": 1}
    PATTERNS = ks.shingles_to_patterns(memo, wildcard="*")
    assert [p.pattern[1:-1] for p in PATTERNS] == [
        "a", "a\\w{1}b", "\\w{1}a", "a\\w{1}"]


# def test9():
#     # rare edge cases
#     k = 5
#     docs = [
#         "Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam ",
#         "nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam ",
#         "erat, sed diam voluptua. At vero eos et accusam et justo duo ",
#         "dolores et ea rebum. Stet clita kasd gubergren, no sea takimata "]
#     # generate all shingles
#     shingled = [ks.shingleseqs_k(doc, k=k) for doc in docs]
#     assert len(shingled) == len(docs)
#     assert len(shingled[0]) == k
#     # run CEWS algorithm
#     db = functools.reduce(lambda x, y: x + Counter(itertools.chain(*y)),
#                           shingled, Counter([]))
#     memo = ks.cews(
#         db, max_wildcards=1, min_samples_leaf=0.0005,
#         threshold=0.9)
#     memo = ks.cews(
#         db, max_wildcards=1, vocab_size=500,
#         min_samples_leaf='auto', threshold=0.9)
#     memo = ks.cews(
#         db, max_wildcards=1, priority='rare',
#         min_samples_leaf=1, threshold=0.9)



import re
from typing import Optional, Dict, List, Union
import functools
import numpy as np


def select_most_frequent_shingles(matches: List[str],
                                  db: Dict[str, int],
                                  min_samples_split: int,
                                  min_samples_leaf: int,
                                  threshold: float):
    """Select the most frequent shingles that matches the wildcard shingle

    Parameters:
    -----------
    matches : List[str]
        A list of shingles from the database (`db`) that matches the current
          wildcard shingle in the `expandshingle` function.

    db: dict
        Database with shingles as `key` and their frequencies/counts as `value`
          Assumptions:
          - Python's `dict` are automatically in alphabetic order but not key
              length, i.e. we still need to filter keys by length initially.
          - The database `db` is immutable.
          Preprocessing: Make sure that each shingle has a sufficient number of
            counts/frequencies, e.g. remove shingles that occur less than 20x.

    min_samples_split: int (Default: 2)
        If the combined frequency of all shingles covered by one wildcard
          shingle (count sum of the regex query results) is less than the
          specified minimum total frequency, then the recursion aborts.

    min_samples_leaf: int (Default: 1)
        Required number of counts in `db` to be added to the memo cached

    threshold: float (Default: 0.80)
        Replace max. `1.0 - threshold` of the least frequent shingles with
          the wildcard shingle.

    Returns:
    --------
    selected_shingles : List[str]
        The selected most frequent shingles

    residual_count : int
        The residual counts (of the unselected shingles) that will be
          assigned to the wildcard shingle in `expandshingle`

    Example:
    --------
        selected_shingles, residual_count = select_most_frequent_shingles(
            matches, db, min_samples_split, min_samples_leaf, threshold)
    """
    # read only matches
    candidates = [item for item in db.items() if item[0] in matches]

    # proceed if there at least 2 candidates
    if len(candidates) == 0:
        return [], 0
    if len(candidates) == 1:
        return [], candidates[0][1]

    # sort by descending frequency
    candidates = sorted(candidates, key=lambda item: -item[1])

    # compute total counts
    total = sum([val for _, val in candidates])

    if total < min_samples_split:
        return [], total

    # find most frequent (`val`) shingles (`key`) up to 90% of all matches
    # always ignore the least frequent shingle and use the wildcard version
    cumpct = 0.0
    cumcnt = 0
    selected = []
    for key, val in candidates[:-1]:
        if val >= min_samples_leaf:
            # abort if the sum of selected shingles reached threshold
            cumpct += val / total
            if cumpct > threshold:
                break
            # select shingle
            cumcnt += val
            selected.append(key)

    # done
    return selected, total - cumcnt


def expandshingle(s: str,
                  db: Dict[str, int],
                  memo: Optional[dict],
                  wildcard: Optional[str] = '\uFFFF',
                  max_wildcards: Optional[int] = 1,
                  min_samples_split: Optional[int] = 2,
                  min_samples_leaf: Optional[int] = 1,
                  threshold: Optional[float] = 0.8):
    """Recursive algorithm to select given k-shingles

    Parameters:
    -----------
    s: str
        A shingle, i.e. a string of text

    db: dict
        Database with shingles as `key` and their frequencies/counts as `value`
          Assumptions:
          - Python's `dict` are automatically in alphabetic order but not key
              length, i.e. we still need to filter keys by length initially.
          - The database `db` is immutable.
          Preprocessing: Make sure that each shingle has a sufficient number of
            counts/frequencies, e.g. remove shingles that occur less than 20x.

    memo: dict (default: {})
        Database for memoization, i.e. `memo[shingle]=count`. In case of the
          wildcarded shingle, the residual counts that are not covered by the
          selected set of shingles (Basically the `1 - p`).
        The keys in the memoization cache are selected shingles of the CEWS
          algorithm.

    wildcard: str  (Default: '\uFFFF')
        An unicode char that is not actually used by any natural language,
          or the text that you analyzing, e.g. U+FFFE or U+FFFF
          See https://en.wikipedia.org/wiki/Specials_(Unicode_block)

    max_wildcards: int (Default: 1)
        If an input string `s` contains more than the specified maximum
          number of wildcard characters, the recursion aborts.

    min_samples_split: int (Default: 2)
        If the combined frequency of all shingles covered by one wildcard
          shingle (count sum of the regex query results) is less than the
          specified minimum total frequency, then the recursion aborts.

    min_samples_leaf: int (Default: 1)
        Required number of counts in `db` to be added to the memo cached

    threshold: float (Default: 0.80)
        Replace max. `1.0 - threshold` of the least frequent shingles with
          the wildcard shingle.

    Returns:
    --------
    memo: dict
        The updated memoization cache.
    """
    # (0a) stop if the shingle reached the maximum number of wildcards
    if s.count(wildcard) >= max_wildcards:
        return memo
    # (0b) abort
    if min_samples_split < 1:
        raise Exception(
            f"min_samples_split={min_samples_split} but must greater equal 1.")

    # (1a) Prefix/Suffix Wildcards
    # - Expand on the left side (prefix wildcard) and right side (suffix w.)
    # - This will lengthen the shingle by 1 character (k+1)
    tmp = []
    tmp.append(f"{wildcard}{s}")   # "_[shingle]"
    tmp.append(f"{s}{wildcard}")   # "[shingle]_"

    # (1b) Infix Wildcard
    # - Check all wildcard variants of a shingle.
    # - This will NOT expand the length of the shingles directly
    n_len = len(s)
    if n_len >= 3:
        # never replace 1st and last with wildcard ("1")
        # and avoid consequtive wildcards at the left and right end
        off_start = 1 + int(s[0] == wildcard)
        off_end = 1 + int(s[-1] == wildcard)
        # create each infix wildcard combination
        for i in range(off_start, n_len - off_end):
            # Find all matches "[s1]_[s2]"
            tmp.append(f"{s[:i]}{wildcard}{s[(i + 1):]}")

    # (2a) Find all matches
    for snew in tmp:
        # memoization trick
        if snew not in memo:
            # regex search
            # reg = re.escape(snew).replace(wildcard, r"\w{1}")
            reg = r"\w{1}".join([re.escape(s) for s in snew.split(wildcard)])
            pat = re.compile(f"^{reg}$")
            matches = list(filter(pat.match, db.keys()))

            # (2b) Find and select the most frequent shingles
            selected_shingles, residual_count = select_most_frequent_shingles(
                matches, db, min_samples_split, min_samples_leaf, threshold)

            # (2c) Assign the counts of the unselected shingles to the new
            #  wildcard-shingle (`residual_count`), store it the database
            #  (`db`) and memoization cache (`memo`), and traverse to the next
            #  knot
            if residual_count >= min_samples_split:
                memo[snew] = residual_count
                memo = expandshingle(
                    snew, db=db, memo=memo,
                    wildcard=wildcard,
                    max_wildcards=max_wildcards,
                    min_samples_split=min_samples_split,
                    min_samples_leaf=min_samples_leaf,
                    threshold=threshold)

                # (2d) Store the selected shingles to the memoization cache
                #  (`memo`), and trigger the next recursion step (traverse
                #  down the tree)
                for key in selected_shingles:
                    if key not in memo:  # memoization trick
                        memo[key] = db[key]
                        memo = expandshingle(
                            key, db=db, memo=memo,
                            wildcard=wildcard,
                            max_wildcards=max_wildcards,
                            min_samples_split=min_samples_split,
                            min_samples_leaf=min_samples_leaf,
                            threshold=threshold)
    # done
    return memo


def cews(db: Dict[str, int],
         memo: Optional[dict] = {},
         wildcard: Optional[str] = '\uFFFF',
         max_wildcards: Optional[int] = 1,
         min_samples_split: Optional[int] = None,
         min_samples_leaf: Optional[int] = 1,
         threshold: Optional[float] = 0.8,
         priority: Optional[str] = 'common',
         vocab_size: Optional[int] = None):
    """Collectively Exhaustive Wildcard Shingling (CEWS)

    Parameters:
    -----------
    db: dict
        Database with shingles as `key` and their frequencies/counts as `value`
          Assumptions:
          - Python's `dict` are automatically in alphabetic order but not key
              length, i.e. we still need to filter keys by length initially.
          - The database `db` is immutable.
          Preprocessing: Make sure that each shingle has a sufficient number of
            counts/frequencies, e.g. remove shingles that occur less than 20x.

    memo: dict
        Add specific shingles to the memoization cache to make sure that these
          part of subword pattern list lateron. These shingles might certain
          keywords, common stopwords, all chars, emojis, abbreviations.
          Call the function as follows:
            import kshingle as ks
            memo = {k: db[k] for k in ["i.e.", "e.g."]}
            memo = ks.cews(db, memo=memo)

    wildcard : str
        An unicode char that is not actually used by any natural language,
          or the text that you analyzing, e.g. U+FFFE or U+FFFF
          See https://en.wikipedia.org/wiki/Specials_(Unicode_block)

    max_wildcards: int (Default: 1)
        If an input string `s` contains more than the specified maximum
          number of wildcard characters, the recursion aborts.

    min_samples_split: int (Default: None)
        If the combined frequency of all shingles covered by one wildcard
          shingle (count sum of the regex query results) is less than the
          specified minimum total frequency, then the recursion aborts.
        If not set, then `min_samples_split = min_samples_leaf / threshold`

    min_samples_leaf: int (Default: 1)
        Required number of counts in `db` to be added to the memo cached.
        `min_samples_leaf` is estimated from `db` if
        - it's a `float` between [0,1]
        - `min_samples_leaf='auto'` and `vocab_size` is set

    threshold: float (Default: 0.80)
        Replace max. `1.0 - threshold` of the least frequent shingles with
          the wildcard shingle.

    priority: str (Default: 'common')
        The order in which to process the shingles.
        - 'common', start with the most frequent shingles. Use this setting
            when specifying high `min_samples_leaf` and/or `vocab_size` to get
            reduce the compute time
        - 'rare', start with less frequent shingles. Use this setting when
            when specifying low `min_samples_leaf=1` and not limiting the
            vocab size.

    vocab_size: int (Default: None)
        Early stopping criteria. The main loop stops if `len(memo) >= vs`.
          The actual memo cache size will be greater equal than `vocab_size`

    Return:
    -------
    memo: dict
        Database for memoization, i.e. `memo[shingle]=count`. In case of the
          wildcarded shingle, the residual counts that are not covered by the
          selected set of shingles (Basically the `1 - p`).
        The keys in the memoization cache are selected shingles of the CEWS
          algorithm.
    """
    # add single-chars to memo automatically
    if len(memo) == 0:
        memo = {k: v for k, v in db.items() if len(k) == 1}

    # sort shingles by frequency
    db_list = list(db.items())
    db_list = sorted(db_list, key=lambda item: -item[1])

    # approximate `min_samples_leaf` with `vocab_size`
    if (min_samples_leaf == 'auto') and (vocab_size is not None):
        idx = min(len(db_list) - 1, vocab_size * 2)
        min_samples_leaf = int(max(1, db_list[idx][1]))

    # if `min_samples_leaf` is float
    if isinstance(min_samples_leaf, float) and (0.0 < min_samples_leaf <= 1.0):
        cnts = np.array([c for _, c in db_list])
        mask = (cnts / np.sum(cnts)) > min_samples_leaf
        if mask.any():
            min_samples_leaf = int(cnts[mask][-1])
        else:
            min_samples_leaf = 1

    # check threshold
    assert threshold is not None
    assert (0.0 < threshold <= 1.0)

    # set `min_samples_split`
    if min_samples_split is None:
        min_samples_split = max(2, int(min_samples_leaf / threshold))

    # only shingles with at least `min_samples_leaf` counts
    if priority == 'rare':
        shingles = [s for s, _ in db_list]
        shingles.reverse()
    elif (priority == 'common') and (vocab_size is not None):
        shingles = [s for s, c in db_list if c >= min_samples_leaf]
    else:  # priority == 'common'
        shingles = [s for s, _ in db_list]

    # loop over all db entries
    for s in shingles:
        memo = expandshingle(
            s, db=db, memo=memo,
            wildcard=wildcard,
            max_wildcards=max_wildcards,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            threshold=threshold)
        # early stopping
        if vocab_size is not None:
            if len(memo) >= vocab_size:
                break

    # done
    return memo


@functools.cmp_to_key
def sort_by_memostats(a, b):
    """ Comparison function list(memo.items())

    Parameters:
    -----------
    a, b : [shingle, num_wild, num_infix, k, count]

    Examples:
    ---------
    wildcard = '\uFFFF'
    MEMOSTATS = [(
        s, len(s.split(wildcard))-1, len(s[1:-1].split(wildcard))-1,
        len(s), c) for s, c in memo.items()]
    MEMOSTATS.sort(key=sort_by_memo)
    shingles = [x[0] for x in MEMOSTATS]
    """
    # (1) Prefer less wilcards
    if a[1] > b[1]:
        return 1
    elif a[1] < b[1]:
        return -1
    else:  # 0: a[1] == b[1]
        # (2) Prefer more infix wildcards
        if a[2] < b[2]:
            return 1
        elif a[2] > b[2]:
            return -1
        else:  # 0: a[2] == b[2]
            # (3) Prefer longer shingle length
            if a[3] < b[3]:
                return 1
            elif a[3] > b[3]:
                return -1
            else:  # 0: same length
                # (4) Prefer more frequent shingles
                if a[4] < b[4]:
                    return 1
                elif a[4] > b[4]:
                    return -1
                else:  # 0: same frequency
                    return 0


def shingles_to_patterns(memo: Dict[str, int],
                         wildcard: Optional[str] = '\uFFFF'
                         ) -> list:  # List[re.Pattern]
    """Convert shingles with wildcards to regex patterns

    Parameters:
    -----------
    memo: dict
        Database for memoization, i.e. `memo[shingle]=count`. In case of the
          wildcarded shingle, the residual counts that are not covered by the
          selected set of shingles (Basically the `1 - p`).
        The keys in the memoization cache are selected shingles of the CEWS
          algorithm.

    wildcard : str
        An unicode char that is not actually used by any natural language,
          or the text that you analyzing, e.g. U+FFFE or U+FFFF
          See https://en.wikipedia.org/wiki/Specials_(Unicode_block)

    Returns:
    --------
    PATTERNS : List[re.Pattern]
        The regex.compile patterns based on the selected shingles in the
          memoization cache.
    """
    # sort memo cache
    if isinstance(memo, dict):
        MEMOSTATS = [(
            s, len(s.split(wildcard)) - 1, len(s[1:-1].split(wildcard)) - 1,
            len(s), c) for s, c in memo.items()]
    elif isinstance(memo, (list, tuple)):
        MEMOSTATS = [(
            s, len(s.split(wildcard)) - 1, len(s[1:-1].split(wildcard)) - 1,
            len(s), 0) for s in memo]
    MEMOSTATS.sort(key=sort_by_memostats)
    shingles = [x[0] for x in MEMOSTATS]
    # convert shingles to regex patterns
    PATTERNS = []
    for shingle in shingles:
        reg = r"\w{1}".join([re.escape(s) for s in shingle.split(wildcard)])
        pat = re.compile(f"^{reg}$")
        PATTERNS.append(pat)
    return PATTERNS


def encode_with_patterns(x: Union[list, str],
                         PATTERNS: list,  # List[re.Pattern]
                         unkid: Optional[int] = None):
    """Encode all elements of x with the regex pattern.

    Parameters:
    -----------
    x : Union[list, str]
        Encoding happens if type(x)==str. If type(x)=list then a recursive
          call on each list element is triggered.

    PATTERNS : List[re.Pattern]
        The regex.compile patterns based on the selected shingles in the
          memoization cache.

    unkid : int
        Index of the UKNOWN token (UNK). It's `unkid=len(PATTERNS)` by default

    Returns:
    --------
    encoded : Union[list, int]
        The IDs refer to the position index in PATTERNS list
    """
    if isinstance(x, str):
        n_pat = len(PATTERNS)
        for i in range(n_pat):
            if PATTERNS[i].match(x):
                return i
        return unkid if unkid else n_pat
    else:
        return [encode_with_patterns(el, PATTERNS, unkid) for el in x]



from typing import List, Optional


def pad_shingle_sequence(seq: List[str],
                         n: int,
                         padding: Optional[str] = 'center',
                         placeholder: Optional[int] = None,
                         evenpad: Optional[str] = 'pre'):
    """Padding of multdimensional shingle sequences

    Parameters:
    -----------
    seq : List[str]
        Sequences with shingles

    n : int
        Length of each shingle

    padding: Optional[str] = 'center'
        Type of padding. 'pre' adds placeholder at the beginning of the
          sequence, 'post' at the end, and 'center' adds placeholder on
          both sides of the sequence.

    placeholder : Optional[int] = None
        The placeholder symbol

    evenpad: Optional[str] = 'pre'
        If padding='center' then sequences with even `n` cannot be perfectly
          centered, i.e. 1 placeholder must be added before (evenpad='pre')
          or after (evenpad='post').
    """
    if padding == 'center':
        # pad left and right
        pads = [placeholder] * ((n - 1) // 2)
        seq = pads + seq + pads
        # pad 1 element if `n` is even
        if (n % 2) == 0:
            if evenpad == 'post':
                seq += [placeholder]
            else:
                seq = [placeholder] + seq
        return seq

    elif padding == 'pre':
        pads = [placeholder] * (n - 1)
        return pads + seq

    elif padding == 'post':
        pads = [placeholder] * (n - 1)
        return seq + pads

    else:
        return seq


def shingleseqs_k(s: str,
                  k: int,
                  padding: Optional[str] = None,
                  placeholder: Optional[int] = None,
                  evenpad: Optional[str] = 'pre'
                  ) -> List[List[str]]:
    """Convert a string to a list of k sequences with k-shingles

    Parameters:
    -----------
    s: str
        The raw string

    k: int
        The parameter must be k>=1

    padding: Optional[str] = None
        Type of padding. 'pre' adds placeholder at the beginning of the
          sequence, 'post' at the end, and 'center' adds placeholder on
          both sides of the sequence.

    placeholder : Optional[int] = None
        The placeholder symbol

    evenpad: Optional[str] = 'pre'
        If padding='center' then sequences with even `n` cannot be perfectly
          centered, i.e. 1 placeholder must be added before (evenpad='pre')
          or after (evenpad='post').

    Returns:
    --------
    List[List[str]]
        A list of k sequences. The 1st sublist contains the chars of the
          string s, e.g. ['a', 'b', 'c'] for s="abc". The k=2 list containd
          2-shingles, e.g. ['ab', 'bc'].

    Example:
    --------
        import kshingle as ks
        shingles = ks.shingleseqs_k("abc", k=2)
        shingles = ks.shingleseqs_k("abc", k=2, padding='center')
    """
    shingles = []
    q = len(s)
    for n in range(1, k + 1):
        seq = [s[i:(i + n)] for i in range(q - n + 1)]
        seq = pad_shingle_sequence(
            seq=seq, n=n, placeholder=placeholder,
            padding=padding, evenpad=evenpad)
        shingles.append(seq)
    return shingles


def shingleseqs_range(s: str,
                      n_min: int,
                      n_max: int,
                      padding: Optional[str] = None,
                      placeholder: Optional[int] = None,
                      evenpad: Optional[str] = 'pre'
                      ) -> List[List[str]]:
    """Convert a string to a list of k sequences with k-shingles

    Parameters:
    -----------
    s: str
        The raw string

    n_min: int
        The lower bound k range [n_min, n_max]

    n_max: int
        The upper bound of the k range [n_min, n_max]

    padding: Optional[str] = None
        Type of padding. 'pre' adds placeholder at the beginning of the
          sequence, 'post' at the end, and 'center' adds placeholder on
          both sides of the sequence.

    placeholder : Optional[int] = None
        The placeholder symbol

    evenpad: Optional[str] = 'pre'
        If padding='center' then sequences with even `n` cannot be perfectly
          centered, i.e. 1 placeholder must be added before (evenpad='pre')
          or after (evenpad='post').

    Returns:
    --------
    List[List[str]]
        A list of k sequences. The k=1 sublist contains the chars of the
          string s, e.g. ['a', 'b', 'c', 'd'] for s="abcd". The k=3 list
          contains 3-shingles, e.g. ['abc', 'bcd'].

    Example:
    --------
        import kshingle as ks
        shingles = ks.shingleseqs_range("abcd", [2, 3])
        shingles = ks.shingleseqs_range("abcd", [2, 3], padding='center')
    """
    # correct wrong inputs
    n_max_ = max(0, n_max)
    n_min_ = max(1, n_min)
    # start to loop
    shingles = []
    q = len(s)
    for n in range(n_min_, n_max_ + 1):
        seq = [s[i:(i + n)] for i in range(q - n + 1)]
        seq = pad_shingle_sequence(
            seq=seq, n=n, placeholder=placeholder,
            padding=padding, evenpad=evenpad)
        shingles.append(seq)
    return shingles


def shingleseqs_list(s: str,
                     klist: List[int],
                     padding: Optional[str] = None,
                     placeholder: Optional[int] = None,
                     evenpad: Optional[str] = 'pre'
                     ) -> List[List[str]]:
    """Convert a string to a list of k sequences with k-shingles

    Parameters:
    -----------
    s: str
        The raw string

    klist: List[int]
        A list of specific k values, e.g. klist=[1, 3, 6]

    padding: Optional[str] = None
        Type of padding. 'pre' adds placeholder at the beginning of the
          sequence, 'post' at the end, and 'center' adds placeholder on
          both sides of the sequence.

    placeholder : Optional[int] = None
        The placeholder symbol

    evenpad: Optional[str] = 'pre'
        If padding='center' then sequences with even `n` cannot be perfectly
          centered, i.e. 1 placeholder must be added before (evenpad='pre')
          or after (evenpad='post').

    Returns:
    --------
    List[List[str]]
        A list of k sequences. The k=1 sublist contains the chars of the
          string s, e.g. ['a', 'b', 'c', 'd'] for s="abcd". The k=3 list
          contains 3-shingles, e.g. ['abc', 'bcd'].

    Example:
    --------
        import kshingle as ks
        shingles = ks.shingleseqs_list("abcd", klist=[1, 3])
        shingles = ks.shingleseqs_list("abcd", klist=[1, 3], padding='center')
    """
    shingles = []
    q = len(s)
    for n in klist:
        if n > 0:
            seq = [s[i:(i + n)] for i in range(q - n + 1)]
            seq = pad_shingle_sequence(
                seq=seq, n=n, placeholder=placeholder,
                padding=padding, evenpad=evenpad)
            shingles.append(seq)
    return shingles
